import numpy as np

from .FRC_lib import radial_profile

def sigmoid(R: float, T: float, S: float):
    '''
    It generates a circularly-symmetric sigmoid function.

    Parameters
    ----------
    R : float
        Radial axis.
    T : float
        Cut-off frequency.
    S : float
        Sigmoid slope.

    Returns
    -------
    np.ndarray
        Sigmoid array. It has the same dimensions of R.

    '''

    return 1 / (1 + np.exp((R - T) / S))


def low_pass(img: np.ndarray, T: float, S: float, data: str = 'real'):
    '''
    It applies a low-pass sigmoidal filter to a 2D image.

    Parameters
    ----------
    img : np.ndarray
        2D image.
    T : float
        Cut-off frequency.
    S : float
        Sigmoid slope.
    data : str, optional
        Domain of the image: It can be 'real' or 'fourier'.
        The default is 'real'.

    Returns
    -------
    img_filt : np.ndarray
        Filtered 2D image, in the domain specified by 'data'.

    '''

    if data == 'real':
        img_fft = np.fft.fftn(img, axes=(0, 1))
        img_fft = np.fft.fftshift(img_fft, axes=(0, 1))
    elif data == 'fourier':
        img_fft = img
    else:
        raise ValueError('data has to be \'real\' or \'fourier\'')

    Nx = np.shape(img_fft)[0]
    Ny = np.shape(img_fft)[1]
    cx = int((Nx + np.mod(Nx, 2)) / 2)
    cy = int((Ny + np.mod(Ny, 2)) / 2)

    x = (np.arange(Nx) - cx) / Nx
    y = (np.arange(Ny) - cy) / Ny

    X, Y = np.meshgrid(x, y)
    R = np.sqrt(X ** 2 + Y ** 2)

    sig = sigmoid(R, T, S)

    img_filt = np.einsum('ij..., ij -> ij...', img_fft, sig)

    if data == 'real':
        img_filt = np.fft.ifftshift(img_filt, axes=(0, 1))
        img_filt = np.fft.ifftn(img_filt, axes=(0, 1))
        img_filt = np.abs(img_filt)

    return img_filt


# %%

def Reorder(dset, inOrder: str, outOrder: str = 'rzxytc'):
    '''
    It reorders a dataset to match the desired order of dimensions.
    If some dimensions are missing, it adds new dimensions.

    Parameters
    ----------
    dset : ndarray
        ISM dataset.
    inOrder : str
        Order of the dimension of the data.
        It can contain any letter of the outOrder string.
    outOrder : str, optional
        Order of the output. The default is 'rzxytc'.

    Returns
    -------
    data : ndarray
        ISM dataset reordered.

    '''

    data = dset.copy()

    Nout = len(outOrder)
    Ndim = len(inOrder)

    if (Ndim < Nout):
        # check where the current dimensions are located
        idx = np.empty( Nout )
        for n, c in enumerate(outOrder):
            idx[n] = np.char.find(inOrder, c)
        idx = idx.astype('int')

        # add missing dimensions
        slices = []
        for i in idx:
            if i == -1:
                slices.append(np.newaxis)
            else:
                slices.append(np.s_[:])

        slices = tuple(slices)
        data = data[slices]

        # reorder final dimensions
        idx2 = np.empty( Ndim )
        for n, c in enumerate(inOrder):
            idx2[n] = np.char.find(outOrder, c)
        idx2 = idx2.astype('int')

        order = idx.copy()
        order[np.where(idx != -1)] = idx2
        order[np.where(idx == -1)] = np.argwhere(idx == -1).flatten()

        data = np.moveaxis(data, np.arange(Nout), order)

    else:
        # check where the dimensions are located
        idx = np.empty( Ndim )
        for n, c in enumerate(inOrder):
            idx[n] = np.char.find(outOrder, c)
        idx = idx.astype('int')

        # remove undesired dimensions
        slices = []
        for i in idx:
            if i == -1:
                slices.append(np.s_[0])
            else:
                slices.append(np.s_[:])

        slices = tuple(slices)
        data = data[slices]

        # reorder remaining dimensions
        order = idx[idx != -1]
        data = np.moveaxis(data, np.arange(Nout), order)

    return data


def CropEdge(dset, npx=10, edges='l', order: str = 'rzxytc'):
    '''
    It crops an ISM dataset along the specified edges of the xy plane.
    
    Parameters
    ----------
    dset : ndarray
        ISM dataset
    npx : int, optional
        Number of pixel to crop from each edge. The default is 10.
    edges : str, optional
        Cropped edges. The possible values are 'l' (left),'r' (right),
        'u' (up), and 'd' (down). Any combination is possible. The default is 'l'.
    order : str, optional
        Order of the dimensions of the dataset The default is 'rzxytc'.

    Returns
    -------
    dset_cropped : ndarray
        ISM dataset cropped

    '''

    default_order = 'rzxytc'

    dset_cropped = Reorder(dset, inOrder = order, outOrder = default_order)

    if 'l' in edges:
        dset_cropped = dset_cropped[..., npx:, :, :, :]

    if 'r' in edges:
        dset_cropped = dset_cropped[..., :-npx, :, :, :]

    if 'u' in edges:
        dset_cropped = dset_cropped[..., :, npx:, :, :]

    if 'd' in edges:
        dset_cropped = dset_cropped[..., :, :-npx, :, :]

    return Reorder(dset_cropped, inOrder = default_order, outOrder = order)


def DownSample(dset, ds: int = 2, order: str = 'rzxytc'):
    '''
    It downsamples an ISM dataset on the xy plane.
    
    Parameters
    ----------
    dset : ndarray
        ISM dataset.
    ds : int, optional
        Downsampling factor. The default is 2.
    order : str, optional
        Order of the dimensions of the dataset The default is 'rzxytc'.
        
    Returns
    -------
    dset_ds : ndarray
        ISM dataset downsampled.

    '''

    default_order = 'rzxytc'

    dset = Reorder(dset, inOrder = order, outOrder = default_order)

    dset_ds = dset[..., ::ds, ::ds, :, :]

    return Reorder(dset_ds, inOrder = default_order, outOrder = order)


def UpSample(dset, us: int = 2, npx: str = 'even', order: str = 'rzxytc'):
    '''
    It upsamples an ISM dataset on the xy plane.

    Parameters
    ----------
    dset : TYPE
        ISM dataset.
    us : int, optional
        Upsampling factor. The default is 2.. The default is 2.
    npx : str, optional
        Parity of the number of pixels on each axis. The default is 'even'.
    order : str, optional
        Order of the dimensions of the dataset The default is 'rzxytc'.

    Returns
    -------
    dset_us : ndarray
        ISM dataset upsampled.

    '''

    default_order = 'rzxytc'

    dset = Reorder(dset, inOrder = order, outOrder = default_order)

    sz = dset.shape

    if npx == 'even':
        sz_us = np.asarray(sz)
        sz_us[2] = sz_us[2] * us
        sz_us[3] = sz_us[3] * us
    elif npx == 'odd':
        sz_us = np.asarray(sz)
        sz_us[2] = sz_us[2] * us - 1
        sz_us[3] = sz_us[3] * us - 1

    dset_us = np.zeros(sz_us)
    dset_us[..., ::us, ::us, :, :] = dset

    return Reorder(dset_us, inOrder = default_order, outOrder = order)


def ArgMaxND(data):
    '''
    It finds the the maximum and the corresponding indeces of a N-dimensional array.

    Parameters
    ----------
    data : ndarray
        N-dimensional array.

    Returns
    -------
    arg : ndarray(int)
        indeces of the maximum.
    mx : float
        maximum value.

    '''

    idx = np.argmax(data)

    mx = np.array(data).ravel()[idx]

    arg = np.unravel_index(idx, np.array(data).shape)

    return arg, mx


def FWHM(x, y):
    '''
    It calculates the Full Width at Half Maximum of a 1D curve.

    Parameters
    ----------
    x : ndarray
        Horizontal axis.
    y : ndarray
        Curve.

    Returns
    -------
    FWHM: float
        Full Width at Half Maximum of the y curve.

    '''

    height = 0.5
    height_half_max = np.max(y) * height
    index_max = np.argmax(y)
    x_low = np.interp(height_half_max, y[:index_max], x[:index_max])
    x_high = np.interp(height_half_max, np.flip(y[index_max:]), np.flip(x[index_max:]))

    return x_high - x_low


def RadialSpectrum(img, pxsize: float = 1, normalize: bool = True):
    '''
    It calculates the radial spectrum of a 2D image.

    Parameters
    ----------
    img : ndarray
        2D image.
    pxsize : float, optional
        Pixel size. The default is 1.
    normalize : bool, optional
        If True, the result is divided by its maximum. The default is True.

    Returns
    -------
    ftR : ndarray
        Radial spectrum.
    space_f : ndarray
        Frequency axis.

    '''

    fft_img = np.fft.fftn(img, axes=[0, 1])
    fft_img = np.abs(np.fft.fftshift(fft_img, axes=[0, 1]))

    sx, sy = fft_img.shape
    c = (sx // 2, sy // 2)

    space_f = np.fft.fftfreq(sx, pxsize)[:c[0]]

    ftR = radial_profile(fft_img, c)

    ftR = ftR[0][:c[0]] / ftR[1][:c[0]]

    ftR = np.real(ftR)

    if normalize == True:
        ftR /= np.max(ftR)

    return ftR, space_f


def fingerprint(dset, volumetric=False):
    """
    Calculate the fingerprint of an ISM dataset.
    The last dimension has to be the spad array channel.

    Parameters
    ----------
    dset : np.array(Nz x Nx x Nx x ... x N*N)
        ISM dataset
    volumetric : bool
        if true, a fingerprint is returned for each axial plane

    Returns
    -------
    Fingerprint : np.array(Nz x N x N)
        Finger print

    """

    N = int(np.sqrt(dset.shape[-1]))

    if volumetric == True:
        Nz = dset.shape[0]
        f = np.empty((Nz, N * N))
        axis = tuple(range(1, dset.ndim - 1))
        f = np.sum(dset, axis=axis)
        f = f.reshape(Nz, N, N)
    else:
        axis = tuple(range(dset.ndim - 1))
        f = np.sum(dset, axis=axis)
        f = f.reshape(N, N)
    return f


import numpy as np
from scipy import pi
import matplotlib.pyplot as plt
import scipy.fft as ft
from statsmodels.nonparametric.smoothers_lowess import lowess
from scipy.optimize import curve_fit

def smooth(x,y):
    """
    Apply smoothing filter based on lowess

    Parameters
    ----------
    x : np.array(N)
        horizontal axis
    y : np.array(N)
        noisy array

    Returns
    -------
    x_interpolated : np.array(100 x N)
        interpolated x-axis
    y_filtered : np.array(100 x N)
        interpolated and smoothed array
    """
    
    x_interp = np.linspace(x[0], x[-1], num = 100*len(x) )
    y_interp = np.interp(x_interp, x, y)
    
    filtered = lowess(y_interp, x_interp, is_sorted=True, frac=0.05, it=0)
    return x_interp, filtered[:,1]

def hann2d(*args):
    """
    Bi-dimensional Hann windowing function

    Parameters
    ----------
    N : int
        Shape of the image (N x M). If only N is provided, a squared
        image (N x N) is assumed.
    M : int
        Shape of the image (N x M)

    Returns
    -------
    W : np.array(N x M)
        Hann window function
    """
    
    if len(args)>1 :
        N, M = args
    else:
        N, = args
        M=N

    x = np.arange(N)
    y = np.arange(M)
    X, Y = np.meshgrid(x,y)
    W = 0.5 * ( 1 - np.cos( (2*pi*X)/(N-1) ) )
    W *= 0.5 * ( 1 - np.cos( (2*pi*Y)/(M-1) ) )
    return W


def radial_profile(data, center):
    """
    Calculation of the radial profile of an image

    Parameters
    ----------
    data : np.array(N x M)
        image
    center : np.array(2)
        indices of the center of the image

    Returns
    -------
    radialprofile : np.array( np.sqrt(N**2 + M**2) )
        sum of the data over the angular coordinate
    nr : int
        number of pixels in the angular bin
    """
    
    y, x = np.indices((data.shape))
    r = np.sqrt((x - center[0])**2 + (y - center[1])**2)
    r = r.astype(int)
    
    tbin = np.bincount(r.ravel(), np.real(data).ravel()).astype(np.complex128)
    tbin += 1j*np.bincount(r.ravel(), np.imag(data).ravel())    
    
    nr = np.bincount(r.ravel())
    radialprofile = tbin

    return radialprofile, nr


def FRC(im1, im2):
    """
    Fourier Ring Correlation. It requires two identical images, differing only
    in the noise content.

    Parameters
    ----------
    im1 : np.array(N x M)
        First input image
    im2 : np.array(N x M)
        Second input image

    Returns
    -------
    FRC : np.array()
        Raw FRC curve
    """
    
    M, N = np.shape(im1)
    cx = int ( ( N + np.mod(N,2) ) / 2)
    cy = int ( ( M + np.mod(M,2) ) / 2)
    center = [cx, cy]
    
    im1 = im1 * hann2d(N,M)
    im2 = im2 * hann2d(N,M)
    
    ft1 = ft.fftshift( ft.fft2( im1 ) )
    ft2 = ft.fftshift( ft.fft2( im2 ) )
    
    num = np.real( radial_profile( ft1*np.conj(ft2) , center)[0] )
    den = np.real( radial_profile( np.abs(ft1)**2, center)[0] )
    den = den * np.real( radial_profile( np.abs(ft2)**2, center)[0] )
    den = np.sqrt( den )
    
    FRC = num / den

    return FRC


def fixed_threshold(frc, y):
    """
    Calculate the treshold for the FRC analysis

    Parameters
    ----------
    frc : np.ndarray
        FRC values
    y : float
        Threshold value

    Returns
    -------
    th : float
        Threshold value
    idx : int
        Index where threshold value is reached
    """
    N = len(frc)
    th = np.ones(N) * y
    try:
        idx = np.argwhere(np.diff(np.sign(frc - y))).flatten()[0]
    except:
        idx = 0
    return th, idx


def nsigma_threshold(k, frc, img, sigma):
    """
    Find the treshold for the FRC analysis

    Parameters
    ----------
    k : np.ndarray
        Frequencies array (N)
    frc : np.ndarray
        FRC value for each k value (N)
    img : np.ndarray()
        Image used to calculate the radial profile
    sigma : int
        Criterium used for the threshold (3 for '3 sigma' criterium, etc.)

    Returns
    -------
    th_interp : float
        Threshold
    idx2 : int
        Index where threshold value is reached
    """
    
    N, M = np.shape(img)
    cx = int ( ( N + np.mod(N,2) ) / 2)
    cy = int ( ( M + np.mod(M,2) ) / 2)
    center = [cx, cy]
    
    nr = radial_profile(img, center)[1]
    
    th = sigma/np.sqrt(nr/2)
    
    k_interp, th_interp = smooth(k, th)
    
    try:
        idx1 = np.argwhere(np.diff(np.sign(frc - th_interp))).flatten()
        idx2 = idx1[1]
    except:
        idx2 = 0
    return th_interp, idx2


def FRC_resolution(I1, I2, px = 1, method = 'fixed', smoothing = 'lowess'):
    """
    Fourier Ring Correlation analysis. It requires two identical images, differing only
    in the noise content, and estimates the resolution from the FRC curve.

    Parameters
    ----------
    I1 : np.array(N x M)
        First input image
    I2 : np.array(N x M)
        Second input image
    px : float
        Pixel size of the images
    method : str
        Threshold criterium. If 'fixed', it uses the 1/7 threshold.
        Other possibilities are '3sigma' and '5sigma'
    smoothing : str
        Smoothing method for the FRC curve. If 'lowess' it smooths and
        interpolates the curve using a lowess algorithm. If 'fit' it fits the
        curve with a sigmoid model and removes high-frequency offset, if present.
        Default is 'lowess'.

    Returns
    -------
    res_um : float
        Estimated resolution, in real units
    k : np.array( np.sqrt(N**2 + M**2) )
        Array of spatial frequencies
    frc : np.array( np.sqrt(N**2 + M**2) )
        Array of raw FRC
    k_interp : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated array of spatial frequencies
    frc_smooth : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated and smoothed FRC curve
    th : np.array( 100 x np.sqrt(N**2 + M**2) )
        Threshold curve
    """
    
    frc = FRC(I1, I2)
    F = len(frc)

    max_kpx = 1 / np.sqrt( 2 )

    kpx = np.linspace(0, max_kpx, F, endpoint = True)
    k = kpx / px

    if smoothing == 'lowess':

        k_interp, frc_smooth = smooth(k, frc) # FRC smoothing

    elif smoothing == 'fit':

        p0 = (1, 1, 10, 0)
        bounds = ( (0,0,0,0), (np.inf, np.inf, np.inf, np.inf) )

        sigmoid_fit = lambda x, a, b, c, d: a / (1 + np.exp((x - b) / c)) + d

        popt, pcov = curve_fit(sigmoid_fit, k[kpx<0.5], frc[kpx<0.5], p0, bounds = bounds)

        amplitude = popt[0]
        offset = popt[-1]

        k_interp = np.linspace(0, max_kpx, F*100, endpoint = True) / px
        frc_smooth = (sigmoid_fit(k_interp, *popt) - offset)/ amplitude
        frc = (frc - offset)/ amplitude

    else:
        raise ValueError('The smoothing parameter has to be "fit" or "lowess".')

    if method == 'fixed':
        th, idx = fixed_threshold(frc_smooth, 1/7)
    elif method == '3sigma':
        th, idx = nsigma_threshold(k, frc_smooth, I1, 3)
    elif method == '5sigma':
        th, idx = nsigma_threshold(k, frc_smooth, I1, 5)

    res_um = (1/k_interp[idx])
    
    return res_um, k, frc, k_interp, frc_smooth, th

def timeFRC(dset, px = 1, method = 'fixed', smoothing = 'lowess'):
    """
    Fourier Ring Correlation analysis. It requires a single dataset with a
    temporal dimension to generate two images using the even and odd indices
    of the time axis. Then, it estimates the resolution using the FRC analysis.

    Parameters
    ----------
    dset : np.ndarray
        dataset (Nx x Ny x Nt)
    px : float
        Pixel size of the images
    method : str
        Threshold criterium. If 'fixed', it uses the 1/7 threshold.
        Other possibilities are '3sigma' and '5sigma'
    smoothing : str
        Smoothing method for the FRC curve. If 'lowess' it smooths and
        interpolates the curve using a lowess algorithm. If 'fit' it fits the
        curve with a sigmoid model and removes high-frequency offset, if present.
        Default is 'lowess'.

    Returns
    -------
    res_um : float
        Estimated resolution, in real units
    k : np.array( np.sqrt(N**2 + M**2) )
        Array of spatial frequencies
    frc : np.array( np.sqrt(N**2 + M**2) )
        Array of raw FRC
    k_interp : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated array of spatial frequencies
    frc_smooth : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated and smoothed FRC curve
    th : np.array( 100 x np.sqrt(N**2 + M**2) )
        Threshold curve
    """
    
    if dset.shape[-1] % 2 == 0:
        img_even = dset[:, :, 0::2].sum(axis = -1)
        img_odd  = dset[:, :, 1::2].sum(axis = -1)
    else:
        img_even = dset[:, :, 0:-1:2].sum(axis = -1)
        img_odd  = dset[:, :, 1::2].sum(axis = -1)
    
    res_um, k, frc, k_interp, frc_smooth, th = FRC_resolution(img_even, img_odd, px = px, method = method, smoothing = smoothing)
    
    return res_um, k, frc, k_interp, frc_smooth, th

def plotFRC(res_um, k, frc, k_interp, frc_smooth, th, fig = None, ax = None):
    """
    Visualization of the results of the FRC curve. The inputs are exactly the
    outputs of FRC_resolution function.

    Parameters
    ----------
    res_um : float
        Estimated resolution, in real units
    k : np.array( np.sqrt(N**2 + M**2) )
        Array of spatial frequencies
    frc : np.array( np.sqrt(N**2 + M**2) )
        Array of raw FRC
    k_interp : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated array of spatial frequencies
    frc_smooth : np.array( 100 x np.sqrt(N**2 + M**2) )
        Interpolated and smoothed FRC curve
    th : np.array( 100 x np.sqrt(N**2 + M**2) )
        Threshold curve

    Returns
    -------
    fig : plt.Figure
        Matplotlib figure.
    ax : plt.axis
        Matplotlib axis.
    """
    
    if fig == None or ax == None:
        fig, ax = plt.subplots()
        
    ax.plot(k, frc, '.', label = 'FRC - raw')
    ax.plot(k_interp, frc_smooth, '-', linewidth = 3, label = 'FRC - smoothed')
    ax.plot(k_interp, th, linewidth = 3, label = 'Threshold')
    
    ax.legend()
    
    idx = (np.abs(k_interp - 1/res_um)).argmin()
    
    ax.plot(k_interp[idx], frc_smooth[idx], 'o', markersize = 6)
    
    k_max = k[-1]*np.sqrt( 2 )/2
    ax.set_xlim( ( 0, k_max ) )
    ax.set_ylim( ( -0.05, 1.05 ) )
    
    ax.set_xlabel(r'k ($\mathregular{\mu m ^{-1}}$)')
    ax.set_ylabel('FRC')
    
    ax.set_title(f'Resolution = {res_um:.3f}' + r' $\mathregular{\mu m}$')
    
    return fig, ax